	processor	16F676
	title	"Servoset2.asm	Issue 1	Rev B	25/02/06"
;**********************************************************************************************
;
;	Setup box for quad servo driver
;
;	
;	Issue	1	Rev b	
;				
;	Controls setup for the servo driver
;	Produces serial data at 9600B (5v level)
;	Sets both travel endstops and speed in both directions
;	Same facilities as the PC setup program
;	Uses switches with resistor chain and A/D to reduce input pins	
;	Uses PIC 16F676 and internal clock at 4MHz
;**********************************************************************************************

;	processor speed  4MHz	Uses internal clock
;	WDT not used
;	transmit baud rate is 9600
;	serial string is a null, a capital letter starting at A and the number
;	from 0 to 255 as a three byte ascii.
;	scans values on switch 1 for servo number 1 to 4
;	scans value on switch 2 for function (pos 0, pos 1, speed 0, speed1)
;	Value on variable pot for coarse position
;	Two push buttons for inch up and inch down
;	Push buttons for run, save and reset modes
;	LED for run indication
;	Working as intended. 16/12/04
;	Would be useful to have the 'reset' for testing each individual servo
;	in both directions. Need another 8 commands and a change in the drivers.
;	Or maybe use a modified run packet with a number over 255?
;	added oscillator calibrate
;	25/02/06  timimg change to RS232 start bit. Now correct

	__CONFIG	h'3FE4' ; PUTimer on, WDT off, BOR on,	internal clock
	#include <p16f676.inc>
;
;******************************************************************
;

;
;	working registers
;




;	other defs
#DEFINE	Inport	PORTA
#DEFINE Outport	PORTC



TX		equ	3	;rs232 tx on PORTC, 3
RUN		equ	0	;run button
SAVE	equ	1	;save button
RSET	equ 2
IN1		equ 4
IN2		equ 5
RLED	equ	4





;
;	end of definitions


	CBLOCK	0x0020		;start of data block in RAM
	State				;state 
	Sersel				;which servo selected
	Funsel				;which function selected
	Runmode				;flag for run mode
	Temp
	Res1				;A/D channel result
	Val					;value from pot.

	Valtmp				;temp for output
	Bitcnt
	Bytcnt
	Count
	Count1
	Nibtmp
	RS232o				;output byte
	EEtemp
	Num0				;ASCII units
	Num1				;ASCII tens
	Num2				;ASCII hundreds
	Inch				;inch value
	Inchflg				;inch toggle
	

	ENDC
;**************************************************************************
;
;	start of program
;
;*************************************************************************
;
	goto	setup		;reset vector
;
	org	04		;interrupt vector not used
	goto	setup		;
;                             
	org	10		;start of prog


;

 






;*******************************************

setup	movlw	0xFF
		bsf		STATUS,RP0		;page 1
		movwf	TRISA			;port a all inputs
		movwf	WPUA			;enable pullups on port a
		movlw	B'10000111'		;port c 0,1,2 are A/D inputs
		movwf	TRISC			
		call	0x3FF			;
		movwf	OSCCAL
		movlw	B'00010000'
		movwf	ADCON1			;set sampling rate
		movlw	B'01110000'
		movwf	ANSEL			;select analog channels
		movlw	1
		movwf	OPTION_REG		;weak pullups, TMR0 int clock / 4
		bcf		STATUS,RP0		;page 0
		movlw	7
		movwf	CMCON			;turn comparator off
		clrf	State
		clrf	Runmode
		clrf	Inch
		clrf	Inchflg
		clrf	PORTC			;all low

;	end of setup

;**********************************************
		
	
main	call	scan
		addwf	PCL,F
		goto	runchk
		goto	run
		goto	save
		goto	reset
		goto	inch1
		goto	inch2

runchk	bcf		Inchflg,0			;button released
		
run1	btfss	Runmode,0		;already in run mode?
		goto	main
	;	movf	Val,W
	;	movwf	Valtmp
		call	inchcal
	
		call	runpkt			;put out a run packet
		goto	main

run		btfsc	Runmode,0		;already in run mode?
		goto	runchk
		bsf		Runmode,0
		bsf		Outport,RLED	;run light on
		call	runpkt
		goto	main
save	bcf		Outport,RLED	;run off
		bcf		Runmode,0
		clrf	Inch
		call	savepkt
		goto	main
reset	bcf		Outport,RLED	;run off
		bcf		Runmode,0
		clrf	Inch
		call	respkt
		goto	main

inch1	btfsc	Inchflg,0		;already on
		goto	run1
		bsf		Inchflg,0		;set button on
		btfss	Inch,7			;is it negative
		goto	inch1a
		incf	Inch,F
	
		goto	run1
inch1a	incf	Inch,F
		btfsc	Inch,7			;gone over 7F?
		decf	Inch,F
		goto	run1

inch2	btfsc	Inchflg,0		;already on
		goto	run1
		bsf		Inchflg,0		;set button on
		btfsc	Inch,7			;is it positive
		goto	inch2a
		decf	Inch,F
		goto	run1
inch2a	decf	Inch,F
		btfss	Inch,7
		incf	Inch,F
		goto	run1
		
		
		
;
;*********************************************************	
;	Subroutines

;	main scanning routine	
scan	bsf		Outport,5
		bcf		Outport,5
		movlw	B'00011001'		;loop
		movwf	ADCON0
		call	del
		bsf		ADCON0,1		;go
wait	btfsc	ADCON0,1
		goto	wait			;conversion done channel 6
		movf	ADRESH,W		;get result
		movwf	Val
		movlw	B'00010001'		
		movwf	ADCON0
		call	del
		bsf		ADCON0,1		;go
wait1	btfsc	ADCON0,1
		goto	wait1
		movf	ADRESH,W
		movwf	Res1
		movlw	B'00010101'		
		movwf	ADCON0
		call	del
		bsf		ADCON0,1		;go channel 5
		clrf	Sersel
		movlw	.43
		subwf	Res1,W			;test which switch position
		btfss	STATUS,C
		goto	serset
		incf	Sersel,F
		movlw	.127
		subwf	Res1,W			;test which switch position
		btfss	STATUS,C
		goto	serset
		incf	Sersel,F
		movlw	.213
		subwf	Res1,W			;test which switch position
		btfss	STATUS,C
		goto	serset
		incf	Sersel,F		
serset	btfsc	ADCON0,1		;next conversion done?
		goto	serset
		movf	ADRESH,W
		movwf	Res1			;answer for channel 5
		clrf	Funsel
		movlw	.43
		subwf	Res1,W			;test which switch position
		btfss	STATUS,C
		goto	funset
		incf	Funsel,F
		movlw	.127
		subwf	Res1,W			;test which switch position
		btfss	STATUS,C
		goto	funset
		incf	Funsel,F		;a speed setting
		movlw	.213
		subwf	Res1,W			;test which switch position
		btfss	STATUS,C
		goto	funset
		incf	Funsel,F
				
funset	movlw	2
		andwf	Funsel,W
		btfsc	STATUS,Z		;is 2 or 3
		goto	funs1
		clrf	Inch
		rrf		Val,F			;so reduce to 0 to 7			
		rrf		Val,F
		rrf		Val,F
		rrf		Val,F
		rrf		Val,F
		movlw	7
		andwf	Val,F
funs1	btfss	Inport,RUN
		retlw	1
		btfss	Inport,SAVE
		retlw	2
		btfss	Inport,RSET
		retlw	3
		btfss	Inport,IN1
		retlw	4
		btfss	Inport,IN2
		retlw	5
		retlw	0		;nothing to do
		


;**********************************************************

;	rs232 output routine (9600 baud)
;	data in rs232o	
;	no handshake
;	low is idle for opto drive

rsout	movlw	8
		movwf	Bitcnt
		movf	RS232o,W
		movwf	Nibtmp		;temp for rs232
		bsf		Outport,TX	;start bit
		movlw	.31			;was .29
		call	dely
		nop
		nop
rsroll	rrf		Nibtmp,f	;next bit		 
		btfsc	STATUS,C
		goto	set1a
		nop
		bsf		Outport,TX	;bit is 0
		goto	rsnext
set1a	bcf		Outport,TX	;bit is 1
		nop
		nop
rsnext	movlw	.29
		call	dely
		nop
		nop
		decfsz	Bitcnt,1
		goto	rsroll
		movlw	1
		call	dely

		bcf		Outport,TX	;stop bit 
		movlw	.29
		call	dely
	
	

	return	

;************************************************

inchcal	btfsc	Inch,7		;Inch is negative?
		goto	inchmin
		movf	Inch,W
		addwf	Val,W
		btfsc	STATUS,C	;over FF?
		goto	corr1
		movwf	Valtmp
		return
corr1	movf	Val,W
		sublw	0xFF
		movwf	Inch
		movlw	0xFF
		movwf	Valtmp
		return
inchmin	movf	Inch,W
		sublw	0
		subwf	Val,W
		btfss	STATUS,C
		goto	corr2
		movwf	Valtmp
		return
corr2	movf	Val,W
		sublw	0
		movwf	Inch
		clrf	Valtmp
		return
		



;***********************************************************

;	convert binary value in Val to three ASCII bytes

conv	clrf	Num2
		clrf	Num1
		clrf	Num0
conv0	movlw	.100
		subwf	Valtmp,F
		btfss	STATUS,C
		goto	conv1
		incf	Num2,F
		goto	conv0
conv1	movlw	.100
		addwf	Valtmp,F
conv2	movlw	.10
		subwf	Valtmp,F
		btfss	STATUS,C
		goto	conv3
		incf	Num1,F
		goto	conv2
conv3	movlw	.10
		addwf	Valtmp,W
conv4	movwf	Num0
		movlw	0x30
		addwf	Num2,F
		addwf	Num1,F
		addwf	Num0,F
		return	
		

;******************************************************

;	Sends one Run packet

runpkt	call	conv		;convert to ASCII
		clrf	RS232o
		bsf		Outport,5	;strobe for test purposes
		bcf		Outport,5
		call	rsout
		movf	Sersel,W
		movwf	Temp
		rlf		Temp,F
		rlf		Temp,F
		movf	Funsel,W
		addwf	Temp,W
		addlw	0x41		;convert to ASCII letter
		movwf	RS232o
		call	rsout
		movf	Num2,W
		movwf	RS232o
		call	rsout
		movf	Num1,W
		movwf	RS232o
		call	rsout
		movf	Num0,W
		movwf	RS232o
		call	rsout
		goto	main

;*******************************************************

;	send 20 save packets

savepkt	movlw	.20
		movwf	Count1
savepk1	clrf	RS232o
		call	rsout		;sync byte
		movlw	"@"
		movwf	RS232o
		call	rsout
		movlw	"0"
		movwf	RS232o
		call	rsout
		movlw	"0"
		movwf	RS232o
		call	rsout
		movlw	"0"
		movwf	RS232o
		call	rsout
		decfsz	Count1,F
		goto	savepk1
		goto	main

;*************************************************

;	20 reset packets

respkt	movlw	.20
		movwf	Count1
respk1	clrf	RS232o
		call	rsout		;sync byte
		movlw	"#"
		movwf	RS232o
		call	rsout
		movlw	"0"
		movwf	RS232o
		call	rsout
		movlw	"0"
		movwf	RS232o
		call	rsout
		movlw	"0"
		movwf	RS232o
		call	rsout
		decfsz	Count1,F
		goto	respk1
		goto	main

;**************************************************	

;	Read EEPROM,  address in W,	result returned in W

eeread	bsf		STATUS,RP0
		movwf	EEADR
		bsf		EECON1,RD
		movf	EEDATA,W
		bcf		STATUS,RP0
		return

;write to EEPROM,	address in W, data in EEtemp

eewrite	bsf		STATUS,RP0
		movwf	EEADR
		movf	EEtemp,W
		movwf	EEDATA
		bsf		EECON1,WREN
		bcf		INTCON,GIE			;*
		movlw	0x55
		movwf	EECON2
		movlw	0xAA
		movwf	EECON2
		bsf		EECON1,WR
		bsf		INTCON,GIE			;*
		bcf		STATUS,RP0
eedone	btfss	PIR1,EEIF			;wait for complete
		goto	eedone
		bcf		PIR1,EEIF
		return

;**********************************************************







;	delay
;
dely	movwf	Count
del1	decfsz	Count,F
	goto	del1
	return 
dely1	movwf	Count
dely1a	nop
	nop
	decfsz	Count,F
	goto	dely1a
	return 

dely2	movlw .15
	movwf	Temp
dely2a	movlw	.255
	call	dely1
	decfsz	Temp,F
	goto	dely2a
	return

dely3	clrf	Temp
dely3a	decfsz	Temp,f
	goto	dely3a
	return

del 	movlw	7		;delay for A/D settling time
		movwf	Temp
dels	decfsz	Temp,F
		goto	dels
		return	

;********************************************************************


	end
